#!/bin/bash

#
# load parameters
#
source config.in


#
# compile the generator of small mults
#
${host_compile} -c mults.f90 
${host_compile} -c multrec_gen.f90 
${host_compile} mults.o multrec_gen.o lib_gen.f90 -o lib_gen.x

#
# directory for the library
#
mkdir -p lib

#
# generate the header with max dimension
#

echo "#define LIBSMV_MAX_RANK    ${dims}" >lib/smv.h

#
# generate the generic caller
#

maxsize=-1
numsize=0

for ((myn=1; myn<=${dims}; myn++))  ; do
#for myn in ${dims_small}
#do
  numsize=$((numsize+1))
  maxsize=`echo "$myn $maxsize" | awk '{if ($1>$2) { print $1 } else { print $2 } }'`
done

#
# generate a translation array for the jump table
#
count=0
eles="(/ (I, I = 0, ${dims}) /)"

if [[ "$data_type" == "1" ]]; then
    filetype="d"
    gemv="DGEMV"
    strdat="REAL(KIND=KIND(0.0D0))"
fi
if [[ "$data_type" == "2" ]]; then
    filetype="s"
    gemv="SGEMV"
    strdat="REAL(KIND=KIND(0.0))"
fi
if [[ "$data_type" == "3" ]]; then
    filetype="z"
    gemv="ZGEMV"
    strdat="COMPLEX(KIND=KIND(0.0D0))"
fi
if [[ "$data_type" == "4" ]]; then
    filetype="c"
    gemv="CGEMV"
    strdat="COMPLEX(KIND=KIND(0.0))"
fi



ta="N"
decl="A(M,M), B(M)"
lds="LDA=M"

file="smv.f90"

cd lib

rm -f ${file}
printf "SUBROUTINE smv(M,A,B,C)\n" >> ${file}
printf " IMPLICIT NONE\n" >> ${file}
printf " INTEGER :: M,LDA\n ${strdat} :: C(M), ${decl}\n" >> ${file}
printf " $strdat, PARAMETER :: one=1\n" >> ${file}
printf " ${lds}\n" >> ${file}

printf " SELECT CASE(M)\n" >> ${file}
for ((mym=1; mym<=${dims}; mym++))
do
    myn=1
    myk=$mym
    printf " CASE($mym)\n " >> ${file}
    printf "   CALL smv_${mym}(A,B,C)\n" >> ${file}
done

printf " CASE DEFAULT\n" >> ${file}
printf "   CALL ${gemv}('%s', M, M, one, A, LDA, B, 1, one, C, 1)\n" $ta >> ${file}
printf " END SELECT\n" >> ${file}
printf "END SUBROUTINE smv" >> ${file}

cd ..

#
# generate list of loop bounds to generate for the small library
#

n=1

postfixes=""
for ((m=1; m<=${dims}; m++))  ; do
 k=$m
 postfixes="$postfixes ${m}"
done

#
# for easy parallelism go via a Makefile
#
rm -f Makefile.lib

(
#
# a two stage approach, first compile in parallel, once done,
# execute in parallel
#
printf "all: archive\n"

printf "driver:\n"
printf "\t cd lib/ ; $target_compile -c ${file} \n\n"

printf "archive: compile driver \n"
printf "\t cd lib/ ; ar -r libsmv.a smv*.o \n\n"

printf "compile: "
for pf in $postfixes ; do
    printf "comp_${pf} "
done
printf "\n\n"

#
# all compile rules
#
#for m in ${dims_small}  ; do
#for n in ${dims_small}  ; do
#for k in ${dims_small}  ; do
for ((m=1; m<=${dims}; m++))  ; do
    printf "comp_${m}: \n"
    printf "\t ./lib_gen.x ${m} 1 ${m} ${transpose_flavor} ${data_type} > lib/smv_${m}.f90\n"
    printf "\t cd lib/ ; $target_compile -c smv_${m}.f90 \n\n"
done #; done ; done

) > Makefile.lib

#
# execute makefile compiling all variants and executing them
#


make -j $tasks -f Makefile.lib all

#
# a final test program. Checking correctness and final performance comparison
#

cat << EOF > test_smv.f90
MODULE WTF
  INTERFACE MYRAND
    MODULE PROCEDURE SMYRAND, DMYRAND, CMYRAND, ZMYRAND
  END INTERFACE
CONTAINS
  SUBROUTINE DMYRAND(A)
    REAL(KIND=KIND(0.0D0)), DIMENSION(:,:) :: A
    REAL(KIND=KIND(0.0)), DIMENSION(SIZE(A,1),SIZE(A,2)) :: Aeq
    CALL RANDOM_NUMBER(Aeq)
    A=Aeq
  END SUBROUTINE
  SUBROUTINE SMYRAND(A)
    REAL(KIND=KIND(0.0)), DIMENSION(:,:) :: A
    REAL(KIND=KIND(0.0)), DIMENSION(SIZE(A,1),SIZE(A,2)) :: Aeq
    CALL RANDOM_NUMBER(Aeq)
    A=Aeq
  END SUBROUTINE
  SUBROUTINE CMYRAND(A)
    COMPLEX(KIND=KIND(0.0)), DIMENSION(:,:) :: A
    REAL(KIND=KIND(0.0)), DIMENSION(SIZE(A,1),SIZE(A,2)) :: Aeq,Beq
    CALL RANDOM_NUMBER(Aeq)
    CALL RANDOM_NUMBER(Beq)
    A=CMPLX(Aeq,Beq,KIND=KIND(0.0))
  END SUBROUTINE
  SUBROUTINE ZMYRAND(A)
    COMPLEX(KIND=KIND(0.0D0)), DIMENSION(:,:) :: A
    REAL(KIND=KIND(0.0)), DIMENSION(SIZE(A,1),SIZE(A,2)) :: Aeq,Beq
    CALL RANDOM_NUMBER(Aeq)
    CALL RANDOM_NUMBER(Beq)
    A=CMPLX(Aeq,Beq,KIND=KIND(0.0D0))
  END SUBROUTINE
END MODULE
SUBROUTINE testit(M)
  USE WTF
  IMPLICIT NONE
  INTEGER :: M

  $strdat :: C1(M), C2(M)
  $strdat :: ${decl}
  $strdat, PARAMETER :: one=1
  INTEGER :: i,LDA, j

  REAL(KIND=KIND(0.0D0)) :: flops,gflop
  REAL :: t1,t2,t3,t4
  INTEGER :: Niter

  flops=2*REAL(M,KIND=KIND(0.0D0))*1*M
  gflop=1000.0D0*1000.0D0*1000.0D0
  ! assume we would like to do 5 Gflop for testing a subroutine
  Niter=MAX(1,CEILING(MIN(10000000.0D0,5*gflop/flops)))
  ${lds}

  DO i=1,10
     CALL MYRAND(A)

     DO j=1,M
        B (j)=i+j+A(1,1)
        C1(j)=i+j+2*A(1,1)
     ENDDO

     C2=C1

     CALL ${gemv}("$ta",M,M,one,A,LDA,B,1,one,C1,1)
     CALL smv(M,A,B,C2)

     IF (MAXVAL(ABS(C2-C1))>100000*EPSILON(REAL(1.0,KIND=KIND(A(1,1))))) THEN
        write(6,*) "Matrix size",M,1,M
        write(6,*) "A=",A
        write(6,*) "B=",B
        write(6,*) "C1=",C1
        write(6,*) "C2=",C2
        write(6,*) "BLAS and smv yield different results : possible compiler bug... do not use the library ;-)"
        STOP
     ENDIF
  ENDDO

  CALL CPU_TIME(t1) 
  DO i=1,Niter
     CALL ${gemv}("$ta",M,M,one,A,LDA,B,1,one,C1,1)
  ENDDO
  CALL CPU_TIME(t2) 

  CALL CPU_TIME(t3)
  DO i=1,Niter
     CALL smv(M,A,B,C2)
  ENDDO
  CALL CPU_TIME(t4)

  WRITE(6,'(A,I5,I5,I5,A,F12.3,A,F12.3,A,F12.3,A)') "Matrix size ",M,1,M, &
        " smv: ",Niter*flops/(t4-t3)/gflop," Gflops. Linked blas: ",Niter*flops/(t2-t1)/gflop,&
        " Gflops. Performance ratio: ",((t2-t1)/(t4-t3))*100,"%"

END SUBROUTINE 

PROGRAM tester
  IMPLICIT NONE

EOF

for ((m=1; m<=${dims}; m++))  ; do
  echo "   CALL testit(${m})" >> test_smv.f90
done

cat << EOF >> test_smv.f90
   ! checking 'peak' performance (and a size likely outside of the library)
   CALL testit(32)
   CALL testit(64)
   CALL testit(128)
   CALL testit(256)
   CALL testit(500)
   CALL testit(512)
   CALL testit(600)
   CALL testit(800)
   CALL testit(1024)
   CALL testit(10000)
END PROGRAM
EOF

#
# compile the benchmarking and testing program for the smv library
#
${target_compile} test_smv.f90 -o test_smv.x -Llib -lsmv ${blas_linking}

#
# run and compile the final test
#
./test_smv.x | tee ./test_smv.out

#
# We're done... protect the user from bad compilers
#
grep 'BLAS and smv yield different results' ./test_smv.out >& /dev/null
if [ "$?" == "0" ]; then
   echo "Library is miscompiled ... removing lib/libsmv.a"
   rm -f lib/libsmv.a
else
   pathhere=`pwd`
   echo "Done... check performance looking at test_smv.out"
   echo "Final library can be linked as -L${pathhere}/lib -lsmv"
fi

